/* gcc -std=c99 -o graph-annealing graph-annealing.c $(sdl-config --cflags --libs) -lGL -lGLU */
/* algorithm from Fruchterman & Reingold 1991 */

#include <assert.h>
#include <time.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include "SDL.h"
#include "SDL_opengl.h"

#define WIDTH 1024
#define HEIGHT 768
#define DEPTH 0

#define EDGE_DENSITY 3.0
#define AUTO_ADD_EDGES 1

#define MAX_TEMPERATURE 0.06f
#define HEAT_FACTOR 1.100f
#define COOL_FACTOR 0.975f
#define ANNEAL_STEP_MSEC 20
#define MIN_DISPLACEMENT 0.00001f
#define MANUAL_COOL_THRESH 0.005f
#define NONEQUAL_VERTEX_REPEL_MULT 1.0f
#define CONNECTED_VERTEX_ATTRACT_MULT 1.0f

#define VERTEX_RADIUS 0.05f
#define MAX_PERTURB (VERTEX_RADIUS * 3.0f)
#define MIN_REPEL_DISTANCE (VERTEX_RADIUS * 0.00001f)
#define EDGE_REPEL 0
#define EDGE_REPEL_CENTER_BIAS_MULT 0.90f
#define EDGE_REPEL_VERTEX_MULT 5.0f
#define EDGE_REPEL_EDGE_MULT 0.2f

#define CIRCLE_STEPS 30
#define FADE_SPEED 5.0f

enum fademode {
	FADE_NONE,
	FADE_IN,
	FADE_OUT,
	FADE_GARBAGE
};

typedef struct graphvertex {
	struct graphvertex *next;
	GLfloat x, y;
	GLfloat dx, dy;
	GLfloat rgba[4];
	enum fademode fademode;
} graphvertex;

typedef struct graphedge {
	struct graphedge *next;
	graphvertex *u, *v;
	enum fademode fademode;
	GLfloat alpha;
} graphedge;

static graphvertex *first_vertex = NULL;
static int vertex_count = 0, fading_vertex_count = 0;
static graphedge *first_edge = NULL;
static int edge_count = 0, fading_edge_count = 0;
static GLfloat temperature = MAX_TEMPERATURE / 10;
static GLfloat bbox[4] = { 0.0f, 0.0f, -1.f, 1.f };

static void
anneal_heat(void)
{
	temperature *= HEAT_FACTOR;
}

static void
anneal_cool(void)
{
	temperature *= COOL_FACTOR;
}

static void
randomize_position(graphvertex *v)
{
	v->x = bbox[0] + (bbox[1] - bbox[0]) * (GLfloat)drand48();
	v->y = bbox[2] + (bbox[3] - bbox[2]) * (GLfloat)drand48();
}

static void
perturb_position(graphvertex *v)
{
	v->x = fminf(bbox[1], fmaxf(bbox[0], v->x + (((drand48()*2.0)-1.0) * MAX_PERTURB)));
	v->y = fminf(bbox[3], fmaxf(bbox[2], v->y + (((drand48()*2.0)-1.0) * MAX_PERTURB)));
}

static void
randomize_color(graphvertex *v)
{
	v->rgba[0] = 0.5f + drand48() * 0.5f;
	v->rgba[1] = 0.5f + drand48() * 0.5f;
	v->rgba[2] = 0.5f + drand48() * 0.5f;
}

static void
randomize_positions(void)
{
	graphvertex *v = first_vertex;
	while (v) {
		randomize_position(v);
		v = v->next;
	}
}

static void
perturb_positions(void)
{
	graphvertex *v = first_vertex;
	while (v) {
		perturb_position(v);
		v = v->next;
	}
}

static void
add_edge(graphvertex *u, graphvertex *v)
{
	graphedge *const e = (graphedge *)malloc(sizeof(graphedge));
	e->next = first_edge;
	e->u = u;
	e->v = v;
	e->alpha = 0.0f;
	e->fademode = FADE_IN;
	first_edge = e;
	edge_count++;
	fading_edge_count++;
}

static void
fade_edge(graphedge *e)
{
	if (e->fademode == FADE_NONE)
		fading_edge_count++;
	e->fademode = FADE_OUT;
}

static void
delete_edges_of(graphvertex *v)
{
	graphedge *e;
	for (e = first_edge; e; e = e->next) {
		if (e->u == v || e->v == v)
			fade_edge(e);
	}
}

static void
delete_random_edge(void)
{
	if (edge_count == 0)
		return;

	long which = lrand48() % edge_count;
	graphedge *e = first_edge;
	while (which--) e = e->next;
	fade_edge(e);
}

static void
add_random_edges(graphvertex *u)
{
	graphvertex *v = first_vertex;
	double const density = fmin((double)EDGE_DENSITY / vertex_count, 0.8);

	while (v) {
		if (u != v && (drand48() < density)) {
			add_edge(u, v);
		}
		v = v->next;
	}
}

static int
edge_between(graphvertex *u, graphvertex *v)
{
	graphedge *e = first_edge;
	while (e) {
		if (e->u == u && e->v == v || e->v == u && e->u == v)
			return 1;
		e = e->next;
	}
	return 0;
}

static void
add_random_edge(void)
{
	if (vertex_count == 0)
		return;
	if (edge_count == (vertex_count * (vertex_count-1)) / 2)
		return;

	graphvertex *u, *v;

	do {
		long i = lrand48() % (vertex_count - 1);
		long j = (lrand48() % (vertex_count - i - 1)) + 1;

		u = first_vertex;
		while (i--) u = u->next;
		v = u;
		while (j--) v = v->next;
	} while (edge_between(u, v));

	add_edge(u, v);
}

static void
push_vertex(void)
{
	graphvertex *const v = (graphvertex *)malloc(sizeof(graphvertex));
	v->next = first_vertex;
	randomize_position(v);
	randomize_color(v);
	v->rgba[3] = 0.0f;
	v->fademode = FADE_IN;
	first_vertex = v;
	vertex_count++;
	fading_vertex_count++;
#if AUTO_ADD_EDGES
	add_random_edges(v);
#endif
}

static void
fade_vertex(graphvertex *v)
{
	if (v->fademode == FADE_NONE)
		fading_vertex_count++;
	v->fademode = FADE_OUT;
}

static void
pop_vertex(void)
{
	if (!first_vertex)
		return;
	fade_vertex(first_vertex);
	delete_edges_of(first_vertex);
}

/* TODO: factor fade stuff */

static void
delete_random_vertex(void)
{
	if (vertex_count == 0)
		return;

	long which = lrand48() % vertex_count;
	graphvertex *v = first_vertex;
	while (which--) v = v->next;
	fade_vertex(v);
	delete_edges_of(v);
}

static void
anneal_step(void)
{
	if (temperature < MIN_DISPLACEMENT || vertex_count == 0)
		return;

	GLfloat const area = (bbox[1] - bbox[0]) * (bbox[3] - bbox[2]);
	GLfloat const k = sqrtf(area / vertex_count);
	GLfloat const k2 = k * k;
	GLfloat dx, dy, dmag;
	graphvertex *u, *v, *w;
	graphedge *e;
	void get_normalized_delta(void) {
		dx = v->x - u->x; dy = v->y - u->y;
		dmag = hypotf(dx, dy);
		dx /= dmag; dy /= dmag;
	}

	for (v = first_vertex; v; v = v->next) {
		v->dx = v->dy = 0.0f;
		for (u = first_vertex; u; u = u->next) {
			if (u != v) {
				/* Nonequal vertexes repel each other. */
				get_normalized_delta();
				GLfloat const repel = k2 / dmag * NONEQUAL_VERTEX_REPEL_MULT;
				v->dx += dx * repel; v->dy += dy * repel;
			}
		}
	}

	for (e = first_edge; e; e = e->next) {
		/* Vertexes connected by an edge attract each other. */
		u = e->u;
		v = e->v;
		get_normalized_delta();
		GLfloat const attract = (dmag * dmag) / k * CONNECTED_VERTEX_ATTRACT_MULT;
		u->dx += dx * attract; u->dy += dy * attract;
		v->dx -= dx * attract; v->dy -= dy * attract;
	}

#if EDGE_REPEL
	for (e = first_edge; e; e = e->next) {
		u = e->u;
		v = e->v;

		GLfloat const e_dx = v->x - u->x;
		GLfloat const e_dy = v->y - u->y;
		GLfloat const e_mag2 = e_dx * e_dx + e_dy * e_dy;

		for (w = first_vertex; w; w = w->next) {
			if (u != w && v != w) {
				/* Edges repel vertices. */
				GLfloat const unnorm_fraction = (w->x - u->x) * e_dx + (w->y - u->y) * e_dy;
				GLfloat const fraction = unnorm_fraction / e_mag2;
				if (fraction <= 0.0f || fraction >= 1.0f) {
					/* A vertex at the end of the edge blocks the edge repulsion. */
					continue;
				}

				/* We bias the vertex position a little towards the center, ad hoc, because
				   the usual case where we want to avoid this is when vertices and edges get pushed
				   all the way to the boundary.
				 */
				GLfloat const opposing_x = u->x + e_dx * fraction;
				GLfloat const opposing_y = u->y + e_dy * fraction;
				GLfloat const opposing_dx = EDGE_REPEL_CENTER_BIAS_MULT * w->x - opposing_x;
				GLfloat const opposing_dy = EDGE_REPEL_CENTER_BIAS_MULT * w->y - opposing_y;
				GLfloat const distance = hypotf(opposing_dx, opposing_dy);
				GLfloat const repel_distance = fmaxf(MIN_REPEL_DISTANCE, distance);
#if 0
				GLfloat const fraction_centered = fraction - 0.5f;
				GLfloat const repel_mult = sqrtf(2.0f * (0.5f - (fraction_centered * fraction_centered)));
#endif
				GLfloat const repel = (k2 / repel_distance);

				w->dx += opposing_dx * repel * EDGE_REPEL_VERTEX_MULT;
				w->dy += opposing_dy * repel * EDGE_REPEL_VERTEX_MULT;
				u->dx -= opposing_dx * repel * EDGE_REPEL_EDGE_MULT;
				u->dy -= opposing_dy * repel * EDGE_REPEL_EDGE_MULT;
				v->dx -= opposing_dx * repel * EDGE_REPEL_EDGE_MULT;
				v->dy -= opposing_dy * repel * EDGE_REPEL_EDGE_MULT;
			}
			
		}
	}
#endif

	for (v = first_vertex; v; v = v->next) {
		dmag = hypotf(v->dx, v->dy);
		GLfloat const wantmag = fminf(dmag, temperature);
		if (wantmag < MIN_DISPLACEMENT)
			continue;
		dx = v->dx / dmag * wantmag;
		dy = v->dy / dmag * wantmag;
		v->x = fminf(bbox[1], fmaxf(bbox[0], v->x + dx));
		v->y = fminf(bbox[3], fmaxf(bbox[2], v->y + dy));
	}

	anneal_cool();
}

static GLuint circle_list = 0;

static void
setup_circle_list(void)
{
	circle_list = glGenLists(1);
	assert(circle_list != 0);

	glNewList(circle_list, GL_COMPILE); {
		GLfloat x_coords[CIRCLE_STEPS], y_coords[CIRCLE_STEPS];
		int i;

		for (i = 0; i < CIRCLE_STEPS; i++) {
			GLfloat const angle = 2.0f * M_PI * i / CIRCLE_STEPS;
			x_coords[i] = cosf(angle) * VERTEX_RADIUS;
			y_coords[i] = sinf(angle) * VERTEX_RADIUS;
		}

		glBegin(GL_LINE_LOOP); {
			for (i = 0; i < CIRCLE_STEPS; i++) {
				glVertex2f(x_coords[i], y_coords[i]);
			}
		}; glEnd();

#if 0
		/* Disabled for now due to interference with vertex fade.
		   Should be done differently anyway.
		 */
		glColor4f(0.f, 0.f, 0.f, .8f);
		glBegin(GL_TRIANGLE_FAN); {
			glVertex2f(0.f, 0.f);
			for (i = 0; i < CIRCLE_STEPS; i++) {
				glVertex2f(x_coords[i] * 0.95f, y_coords[i] * 0.95f);
			}
		}; glEnd();
#endif
	}; glEndList();
}

static void
draw_boundary(void)
{
	glColor4f(.3f, .3f, .3f, 1.f);
	glBegin(GL_LINE_LOOP); {
		glVertex2f(bbox[0] - VERTEX_RADIUS, bbox[2] - VERTEX_RADIUS);
		glVertex2f(bbox[0] - VERTEX_RADIUS, bbox[3] + VERTEX_RADIUS);
		glVertex2f(bbox[1] + VERTEX_RADIUS, bbox[3] + VERTEX_RADIUS);
		glVertex2f(bbox[1] + VERTEX_RADIUS, bbox[2] - VERTEX_RADIUS);
	}; glEnd();
}

static void
draw_vertices(void)
{
	graphvertex *v;

	for (v = first_vertex; v; v = v->next) {
		glPushMatrix();
		glTranslatef(v->x, v->y, 0.f);
		glColor4fv(v->rgba);
		glCallList(circle_list);
		glPopMatrix();
	}
}

static void
draw_edges(void)
{
	graphedge *e;

	glBegin(GL_LINES); {
		for (e = first_edge; e; e = e->next) {
			GLfloat xu = e->u->x, yu = e->u->y;
			GLfloat xv = e->v->x, yv = e->v->y;
			GLfloat const dx = (xv - xu), dy = (yv - yu);
			GLfloat const offmult = VERTEX_RADIUS / hypotf(dx, dy);
			xu += dx * offmult; yu += dy * offmult;
			xv -= dx * offmult; yv -= dy * offmult;

			GLfloat rgba[4];

			memcpy(rgba, e->u->rgba, sizeof(GLfloat) * 3);
			rgba[3] = fminf(e->u->rgba[3], e->alpha);
			glColor4fv(rgba);
			glVertex2f(xu, yu);

			memcpy(rgba, e->v->rgba, sizeof(GLfloat) * 3);
			rgba[3] = fminf(e->v->rgba[3], e->alpha);
			glColor4fv(rgba);
			glVertex2f(xv, yv);
		}
	}; glEnd();
}

static void
draw_graph(void)
{
	glPushMatrix(); {
		GLfloat const scale_x = (bbox[1] - bbox[0]) / (bbox[1] - bbox[0] + VERTEX_RADIUS);
		GLfloat const scale_y = (bbox[3] - bbox[2]) / (bbox[3] - bbox[2] + VERTEX_RADIUS);
		GLfloat const scale = fminf(scale_x, scale_y) * 0.9f;
		glScalef(scale, scale, 1.f);
		draw_boundary();
		draw_edges();
		draw_vertices();
	}; glPopMatrix();
}

static void
fade_edges(GLfloat dt)
{
	graphedge *e;
	for (e = first_edge; e; e = e->next) {
		switch (e->fademode) {
		    case FADE_IN:
			e->alpha = fminf(1.0f, e->alpha + FADE_SPEED * dt);
			if (e->alpha >= 1.0f) {
				e->fademode = FADE_NONE;
				fading_edge_count--;
			}
			break;

		    case FADE_OUT:
			e->alpha = fmaxf(0.0f, e->alpha - FADE_SPEED * dt);
			if (e->alpha <= 0.0f) {
				e->fademode = FADE_GARBAGE;
				fading_edge_count--;
			}
			break;
		}
	}
}

static void
fade_vertices(GLfloat dt)
{
	/* Grr, duplication, buggo. */
	graphvertex *v;
	for (v = first_vertex; v; v = v->next) {
		switch (v->fademode) {
		    case FADE_IN:
			v->rgba[3] = fminf(1.0f, v->rgba[3] + FADE_SPEED * dt);
			if (v->rgba[3] >= 1.0f) {
				v->fademode = FADE_NONE;
				fading_vertex_count--;
			}
			break;

		    case FADE_OUT:
			v->rgba[3] = fmaxf(0.0f, v->rgba[3] - FADE_SPEED * dt);
			if (v->rgba[3] <= 0.0f) {
				v->fademode = FADE_GARBAGE;
				fading_vertex_count--;
			}
			break;
		}
	}
}

static void
gc_edges(void)
{
	graphedge *e = first_edge, **last_next = &first_edge;
	while (e) {
		graphedge *const next = e->next;
		if (e->fademode == FADE_GARBAGE) {
			*last_next = e->next;
			free(e);
			edge_count--;
		} else {
			last_next = &(e->next);
		}
		e = next;
	}
}

static void
gc_vertices(void)
{
	/* Hrm, duplication, buggo. */
	graphvertex *v = first_vertex, **last_next = &first_vertex;
	while (v) {
		graphvertex *const next = v->next;
		if (v->fademode == FADE_GARBAGE) {
			*last_next = v->next;
			free(v);
			vertex_count--;
		} else {
			last_next = &(v->next);
		}
		v = next;
	}
}

static void
fade_step(GLfloat dt)
{
	fade_edges(dt);
	fade_vertices(dt);
	gc_edges();
	gc_vertices();
}

static void
print_docs(void)
{
	fputs("Graph annealing simulator\n"
	      "-------------------------\n"
	      "  q - quit, a - add vertex, d - pop vertex, D - delete random vertex\n"
	      "  r - perturb positions, R - randomize positions, p = heat, P = cool\n"
	      "  x - add random edge, X - delete random edge\n"
	      "  <space> - pause/unpause annealing\n",
	      stdout);
}

int
main(int argc, char *argv[])
{
	assert(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) == 0);
	assert(SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1) == 0);
	assert(SDL_SetVideoMode(WIDTH, HEIGHT, DEPTH, SDL_OPENGL | SDL_FULLSCREEN) != NULL);
	SDL_WM_SetCaption("Graph Annealing Demo", "Graph Annealing Demo");
	SDL_ShowCursor(SDL_DISABLE);
	SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL / 2);
	srand48(time(NULL));
	setup_circle_list();
	print_docs();
	bbox[0] = -(GLfloat)WIDTH / (GLfloat)HEIGHT;
	bbox[1] = +(GLfloat)WIDTH / (GLfloat)HEIGHT;

	/* TODO: make bbox dynamic */
	glViewport(0, 0, WIDTH, HEIGHT);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(bbox[0], bbox[1], bbox[2], bbox[3]);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glDisable(GL_LIGHTING); glDisable(GL_TEXTURE_2D);
	glDisable(GL_TEXTURE_1D); glDisable(GL_ALPHA_TEST); glDisable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glClearColor(0.f, 0.f, 0.f, 0.f);

	int do_anneal = 1, do_heat = 0, do_cool = 0;
	Uint32 last_frame = SDL_GetTicks();
	Uint32 last_anneal = last_frame;

	for (;;) {
		SDL_Event ev;

		while (SDL_PollEvent(&ev)) {
			if (ev.type == SDL_QUIT) {
				goto quit;
			} else if (ev.type == SDL_KEYDOWN) {
				switch (ev.key.keysym.sym) {
				    case SDLK_q:
					goto quit;
				    case SDLK_a:
					push_vertex();
					break;
				    case SDLK_d:
					if (ev.key.keysym.mod & KMOD_SHIFT)
						delete_random_vertex();
					else
						pop_vertex();
					break;
				    case SDLK_r:
					if (ev.key.keysym.mod & KMOD_SHIFT)
						randomize_positions();
					else
						perturb_positions();
					break;
				    case SDLK_p:
					if (ev.key.keysym.mod & KMOD_SHIFT)
						do_cool = 1;
					else
						do_heat = 1;
					break;
				    case SDLK_SPACE:
					do_anneal = !do_anneal;
					break;
				    case SDLK_x:
					if (ev.key.keysym.mod & KMOD_SHIFT)
						delete_random_edge();
					else
						add_random_edge();
					break;
				}
			}
		}

		Uint32 const now = SDL_GetTicks();

		int const do_fade = fading_vertex_count || fading_edge_count;
		if (do_fade)
			fade_step((GLfloat)(now - last_frame) / 1000.0f);

		if (do_heat) {
			if (temperature >= MAX_TEMPERATURE)
				do_heat = 0;
			else
				anneal_heat();
		}

		if (do_cool) {
			if (temperature <= MANUAL_COOL_THRESH)
				do_cool = 0;
			else
				anneal_cool();
		}

		if (do_anneal && now - last_anneal >= ANNEAL_STEP_MSEC) {
			anneal_step();
			last_anneal = now;
		}

		glClear(GL_COLOR_BUFFER_BIT);
		draw_graph();

		last_frame = now;
		SDL_GL_SwapBuffers();
		SDL_Delay(5);
	}

quit:
	SDL_Quit();
	return 0;
}
