/* gcc -std=c99 -ffast-math -o gravity gravity.c $(sdl-config --cflags --libs) -lGL -lGLU */
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include <malloc.h>
#include <errno.h>
#include <stdarg.h>
#include "SDL.h"
#include "SDL_opengl.h"
#include <math.h>
#include <time.h>
#include <assert.h>

static void 
die(char const *fmt, ...)
{
	va_list va;
	va_start(va, fmt);
	vfprintf(stderr, fmt, va);
	va_end(va);
	exit(255);
}

static void *xmalloc(size_t size)
{
	void *ptr;
	if (size == 0)
		return 0;
	ptr = malloc(size);
	if (!ptr)
		die("Out of memory!\n");
	return ptr;
}

static void xfree(void *ptr)
{
	if (ptr)
		free(ptr);
}

typedef struct {
	double pos[2];
	double mass;
	double velocity[2];
	double accel[2];
	int dead;
} mobile_s;

typedef struct {
	double pos[2];
	double mass;
	double radius;
} fixed_s;

static unsigned const WIDTH = 1024;
static unsigned const HEIGHT = 768;
static unsigned const N_FIXED = 10;
static unsigned const N_MOBILE = 250;
static double const SEPARATION = 0.0001;
static double const START_SPEED = 0.4;
static double const GRAVITY = 2e-3;
static double const DT = 2e-3;

static GLuint
make_texture(void)
{
	static unsigned const SIZE = 256;
	static GLubyte const FROM[3] = { 255, 0, 0 };
	static GLubyte const TO[3] = { 0, 255, 0 };

	GLubyte *pixels = xmalloc(SIZE * 3);
	for (unsigned i = 0; i < SIZE; i++) {
		double const frac = (double)i / (double)(SIZE-1);
		pixels[i*3+0] = (GLubyte)rint(64.0 + 50.0 * sin((double)frac * 2.0 * M_PI));
		pixels[i*3+1] = (GLubyte)rint(64.0 + 128.0 * frac);
		pixels[i*3+2] = (GLubyte)rint(70.0 + 10.0 * frac
					      + 35.0 * cos((double)frac * 2.5 * M_PI));
	}

	GLuint t;
	glGenTextures(1, &t);

	glPushAttrib(GL_TEXTURE_BIT);
	glBindTexture(GL_TEXTURE_1D, t);
	glTexImage1D(GL_TEXTURE_1D, 0, GL_RGB, SIZE, 0, GL_RGB, GL_UNSIGNED_BYTE, pixels);
	glTexEnvi(GL_TEXTURE_1D, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glPopAttrib();

	free(pixels);
	return t;
}

static GLuint
make_circle_list(void)
{
	static unsigned const N_STEPS = 20;
	GLuint c = glGenLists(1);

	glNewList(c, GL_COMPILE);
	glBegin(GL_LINE_LOOP);
	for (unsigned i = 0; i < N_STEPS; i++) {
		double const angle = ((double)i / (double)N_STEPS) * M_PI * 2 + 0.01;
		glVertex2d(cos(angle), sin(angle));
	}
	glEnd();
	glEndList();

	return c;
}

static fixed_s *
make_fixed(void)
{
	fixed_s *const fs = xmalloc(sizeof(fixed_s) * N_FIXED);
	fixed_s *const fs_end = fs + N_FIXED;
	for (fixed_s *f = fs; f < fs_end; f++) {
		double const mass = (drand48() * 14.0) + 5.0;
		f->mass = mass;
		f->radius = mass * 0.001;
		f->pos[0] = (drand48() * (1.0 - 2*f->radius)) + f->radius;
		f->pos[1] = (drand48() * (1.0 - 2*f->radius)) + f->radius;
		f->mass = (drand48() * 14.0) + 5.0;
	}

	return fs;
}

static mobile_s *
make_mobile(void)
{
	mobile_s *const ms = xmalloc(sizeof(mobile_s) * N_MOBILE);
	mobile_s *const ms_end = ms + N_MOBILE;

	double const mass = 0.5;
	double const start[2] = { drand48(), drand48() };
	double sep_angle = drand48() * 2 * M_PI;
	do {
		double const end[2] = { start[0] + cos(sep_angle) * SEPARATION * (N_MOBILE - 1),
					start[1] + sin(sep_angle) * SEPARATION * (N_MOBILE - 1) };
		if (end[0] >= 0.0 && end[0] < 1.0 && end[1] >= 0.0 && end[1] < 1.0)
			break;
		sep_angle = drand48() * 2 * M_PI;
	} while (1);

	double const delta[2] = { cos(sep_angle) * SEPARATION,
				  sin(sep_angle) * SEPARATION };
	double const vel_angle = drand48() * 2 * M_PI;
	double const start_velocity[2] = { cos(vel_angle) * START_SPEED,
					   sin(vel_angle) * START_SPEED };
	for (mobile_s *m = ms; m < ms_end; m++) {
		m->velocity[0] = start_velocity[0];
		m->velocity[1] = start_velocity[1];
		m->mass = mass;
		m->dead = 0;
		m->pos[0] = start[0] + delta[0] * (m - ms);
		m->pos[1] = start[1] + delta[1] * (m - ms);
	}

	return ms;
}

static void 
draw_fixed(fixed_s *fs, GLuint shape)
{
	return ;

	fixed_s *const fs_end = fs + N_FIXED;
	glColor3f(0.5f, 0.5f, 0.5f);

	for (fixed_s *f = fs; f < fs_end; f++) {
		glPushMatrix();
		glTranslated(f->pos[0], f->pos[1], 0.0);
		glScaled(f->radius, f->radius, f->radius);
		glCallList(shape);
		glPopMatrix();
	}
}

static void 
draw_mobile(mobile_s *ms, GLuint texture)
{
	mobile_s *const ms_end = ms + N_MOBILE;

	glPushAttrib(GL_ENABLE_BIT | GL_TEXTURE_BIT);
	glEnable(GL_TEXTURE_1D);
	glBindTexture(GL_TEXTURE_1D, texture);
	glColor3f(1.0f, 1.0f, 1.0f);

	GLfloat tc = 0.f;
	GLfloat delta_tc = 1.0f / (GLfloat)(N_MOBILE - 1);

/*
	glBegin(GL_LINE_STRIP);
	for (mobile_s *m = ms; m < ms_end; m++) {
		glTexCoord1f(tc);
		if (!m->dead) glVertex2d(m->pos[0], m->pos[1]);
		tc += delta_tc;
	}
	glEnd();
*/

	tc = 0.f;
	glPointSize(3.0f);
	glBegin(GL_POINTS);
	for (mobile_s *m = ms; m < ms_end; m++) {
		glTexCoord1f(tc);
		if (/* 1 || m->dead */ !m->dead) glVertex2d(m->pos[0], m->pos[1]);
		tc += delta_tc;
	}
	glEnd();

	glDisable(GL_TEXTURE_1D);
	glPointSize(4.0f);
	glBegin(GL_POINTS);
	for (mobile_s *m = ms; m < ms_end; m++) {
		if (m->dead >= 1 && m->dead <= 100) {
			double const speed = hypot(m->velocity[0], m->velocity[1]);
			double const gray = pow(atan(speed) / (M_PI/2), 0.25) * (100.0 - (double)m->dead) / 100.0;
			glColor3d(gray, gray, gray);
			glVertex2d(m->pos[0], m->pos[1]);
			m->dead++;
		}
	}
	glEnd();

	glPopAttrib();
}

static void 
affect(mobile_s *target, fixed_s *source)
{
	double const delta[2] = { source->pos[0] - target->pos[0],
				  source->pos[1] - target->pos[1] };
	double const distance2 = delta[0] * delta[0] + delta[1] * delta[1];
	double const distance = sqrt(distance2);
	double const norm_delta[2] = { delta[0] / distance, delta[1] / distance };
	double const mag_accel = GRAVITY * source->mass / distance2;
	double const accel[2] = { mag_accel * norm_delta[0], mag_accel * norm_delta[1] };
	target->accel[0] += accel[0];
	target->accel[1] += accel[1];
}

static void 
move_mobile(mobile_s *ms, int *n_dead, fixed_s *fs)
{
	mobile_s *const ms_end = ms + N_MOBILE;
	fixed_s *const fs_end = fs + N_FIXED;

	for (mobile_s *m = ms; m < ms_end; m++) {
		if (m->dead) continue;
		m->accel[0] = m->accel[1] = 0.0;
		for (fixed_s *f = fs; f < fs_end; f++)
			affect(m, f);
	}

	for (mobile_s *m = ms; m < ms_end; m++) {
		if (m->dead) continue;
		m->velocity[0] += m->accel[0] * DT;
		m->velocity[1] += m->accel[1] * DT;
		m->pos[0] += m->velocity[0] * DT;
		m->pos[1] += m->velocity[1] * DT;
		if (m->pos[0] < 0.0 || m->pos[0] > 1.0)
			m->velocity[0] = -m->velocity[0];
		if (m->pos[1] < 0.0 || m->pos[1] > 1.0)
			m->velocity[1] = -m->velocity[1];

		for (fixed_s *f = fs; f < fs_end; f++) {
			if (hypot(m->pos[0] - f->pos[0], m->pos[1] - f->pos[1])
			    < f->radius) {
				m->dead = 1;
				(*n_dead)++;
			}
		}
	}
}

static void 
capture(void)
{
	static GLubyte *pixels = NULL;
	if (pixels == NULL)
		pixels = xmalloc(sizeof(GLubyte) * WIDTH * HEIGHT * 3);

	glReadPixels(0, 0, WIDTH, HEIGHT, GL_RGB, GL_UNSIGNED_BYTE, pixels);
	for (unsigned i = 0; i < HEIGHT/2; i++) {
		GLubyte *ptr0 = pixels + WIDTH * 3 * i;
		GLubyte *ptr1 = pixels + WIDTH * 3 * (HEIGHT - 1 - i);
		for (unsigned j = WIDTH * 3; j != 0; j--, ptr0++, ptr1++) {
			GLubyte const tmp = *ptr0;
			*ptr0 = *ptr1;
			*ptr1 = tmp;
		}
	}

	write(3, pixels, sizeof(GLubyte) * WIDTH * HEIGHT * 3);
}

int
main(int argc, char *argv[])
{
	int err;
	SDL_Surface *screen;

	long const seed = (argc > 1)? strtol(argv[1], NULL, 10) : time(NULL);
	srand48(seed);
	printf("Random seed: %ld\n", seed);

	int const writing_video = (getenv("write_video") != NULL);

	err = SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER);
	if (err)
		die("Can't initialize SDL: %s\n", SDL_GetError());

	err = SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	if (err)
		die("Can't set OpenGL doublebuffering: %s\n", SDL_GetError());

	screen = SDL_SetVideoMode(WIDTH, HEIGHT, 0, SDL_OPENGL | SDL_FULLSCREEN);
	if (!screen)
		die("Can't set video mode: %s\n", SDL_GetError());
	SDL_ShowCursor(SDL_DISABLE);

	/* BUGGO */
	double const aspect = (double)WIDTH / (double)HEIGHT;
	assert(aspect > 1.0); 	/* Buggo! */
	glViewport(0, 0, WIDTH, HEIGHT);
	glMatrixMode(GL_PROJECTION);
	gluOrtho2D(-(aspect - 1.0)/2, 1.0 + (aspect - 1.0)/2, 0.0, 1.0);
	glMatrixMode(GL_MODELVIEW); glLoadIdentity();

	GLuint const t = make_texture();
	GLuint const c = make_circle_list();
	fixed_s *fs = make_fixed();
	mobile_s *ms = make_mobile();
	int n_dead = 0;

	for (;;) {
		SDL_Event ev;

		while (SDL_PollEvent(&ev)) {
			switch (ev.type) {
			    case SDL_QUIT:
				goto quit;
			    case SDL_KEYDOWN:
				if (ev.key.keysym.sym == SDLK_q) goto quit;
				free(fs);
				free(ms);
				fs = make_fixed();
				ms = make_mobile();
				n_dead = 0;
				break;
			}
		}

		glClearColor(0.f, 0.f, 0.f, 0.f);
		glClear(GL_COLOR_BUFFER_BIT);
		glColor3f(0.2f, 0.2f, 0.2f);
		glBegin(GL_LINE_LOOP);
		glVertex2f(0.f, 0.f);
		glVertex2f(0.f, 1.f);
		glVertex2f(1.f, 1.f);
		glVertex2f(1.f, 0.f);
		glEnd();

		draw_fixed(fs, c);
		draw_mobile(ms, t);
		if (writing_video) capture();
		move_mobile(ms, &n_dead, fs);

		SDL_GL_SwapBuffers();
		SDL_Delay(1);
	}

quit:
	SDL_Quit();
	return 0;
}
