/* gcc -std=c99 -ffast-math -W -Wall -o penrose-sdl penrose-sdl.c $(sdl-config --cflags --libs) -lGL -lGLU */

/* References:
   
     E. A. Lord, K. Ramakrishnan and S. Ranganathan, "An algorithm for
     generating quasiperiodic patterns and their approximants", Bulletin
     of Materials Science, Vol. 23, No. 2, April 2000, pp. 119--123,
     Indian Academy of Sciences.

 */

#include <stdlib.h>
#include <math.h>
#include <sys/types.h>
#include <stdio.h>
#include <assert.h>
#include <time.h>
#include "SDL.h"
#include "SDL_opengl.h"

typedef struct vertex {
	/* BUGGO: stupid seen-vertex mechanism */
	struct vertex *next;
	int *coord;
} vertex_s;

static int const WIDTH = 1024;
static int const HEIGHT = 768;

enum draw_type {
	DRAW_QUAD,
	DRAW_ELLIPSE,
	DRAW_DUAL_VCOLOR,
	DRAW_DUAL_PCOLOR,
};

typedef struct qc_state {
	enum draw_type draw_type;
	int use_pcolor;

	GLuint solidc, wirec;

	vertex_s *seen_vertices;	/* Linked list of traversed vertices */

	unsigned n_hyper;		/* Number of dimensions of the hypercubic lattice. */
	unsigned n_proj;		/* Number of output projection dimensions. */
	unsigned n_rest;		/* Number of non-projected dimensions. */
	/* Invariant: n_hyper = n_proj + n_rest. */
	/* Invariant: n_proj < n_hyper.  Corollary: n_rest < n_hyper. */
	/* Invariant: n_hyper < (CHAR_BIT * sizeof(unsigned)). */

	double *rotation;               /* Rotation matrix, of [n_hyper][n_hyper]. */
	/* Invariant: for all rows A, B of rotation, A is orthogonal to B. */
	/* Invariant: for all rows R of rotation, magnitude(R) = 1. */

	double *offset;			/* Offset vector, of [n_hyper]. */
	/* The offset is applied before rotation. */

	unsigned n_bounds;		/* Number of bound plane pairs. */
	/* Invariant: n_bounds = choose(n_hyper, n_rest - 1). */

	double *bounds;			/* Set of bound plane vectors, of [n_bounds][n_rest]. */
	/* To be inside the bounds, a point's perpendicular-space subvector must yield a number in [-1, 1]
	   when dotted with each bound plane vector. */
} qc_state_s;

static void *
xmalloc(size_t size)
{
	void *const ptr = malloc(size);
	if (!ptr) {
		fputs("quasicrystal: out of memory\n", stderr);
		exit(111);
	}

	return ptr;
}

#define xnew(typ, cnt) ((typ *)xmalloc(sizeof(typ) * (cnt)))

static inline unsigned 
find_biggest_row(unsigned n, double *matrix /* [n][n] */, unsigned from /* < n */)
{
	unsigned max_row = from;
	double max_abs = fabs(matrix[from * n + from]);
	unsigned row;

	/* Find the row with the highest absolute value in this column. */
	for (row = from+1; row < n; row++) {
		double const this_abs = fabs(matrix[row * n + from]);
		if (this_abs > max_abs) {
			max_row = row;
			max_abs = this_abs;
		}
	}

	return max_row;
}

static inline void 
swap_rows(unsigned n, double *matrix /* [n][n] */, double *vector /* [n] */,
	  unsigned i /* < n */, unsigned j /* < n */)
{
	double *const row_i = &matrix[i * n];
	double *const row_j = &matrix[j * n];
	unsigned k;
	double tmp;

	for (k = 0; k < n; k++) {
		tmp = row_i[k];
		row_i[k] = row_j[k];
		row_j[k] = tmp;
	}

	tmp = vector[i];
	vector[i] = vector[j];
	vector[j] = tmp;
}

static inline void 
divide_row(unsigned n, double *matrix /* [n][n] */, double *vector /* [n] */,
	   unsigned i /* < n */, double divisor)
{
	double *const row = &matrix[i * n];
	unsigned k;

	for (k = 0; k < n; k++) row[k] /= divisor;
	vector[i] /= divisor;
}

static inline void 
add_multiplied_row(unsigned n, double *matrix /* [n][n] */, double *vector /* [n] */,
		   unsigned i /* < n */, unsigned j /* < n */, double factor)
{
	double *const target = &matrix[i * n];
	double const *const source = &matrix[j * n];
	unsigned k;

	for (k = 0; k < n; k++) target[k] += source[k] * factor;
	vector[i] += vector[j] * factor;
}

static void 
print_linear(unsigned n, double *matrix /* [n][n] */, double *vector /* [n] */)
{
	unsigned i, j;

	for (i = 0; i < n; i++) {
		printf("[");
		for (j = 0; j < n; j++) {
			printf(" % 0.3f", matrix[i * n + j]);
		}

		printf(" | % 0.3f ]\n", vector[i]);
	}
}

static void 
solve_linear(unsigned n, double *matrix /* [n][n] */, double *vector /* [n] */)
{
	/* Gaussian elimination.  This probably doesn't handle all the cases, sigh. */
	unsigned i, j;
	
	for (i = 0; i < n; i++) {
		swap_rows(n, matrix, vector, i, find_biggest_row(n, matrix, i));
		divide_row(n, matrix, vector, i, matrix[i * n + i]);

		for (j = 0; j < n; j++) {
			if (j == i) continue;
			add_multiplied_row(n, matrix, vector, j, i, -matrix[j * n + i]);
		}
	}
}

static inline unsigned long
choose(unsigned long n, unsigned long k)
{
	unsigned long acc = 1, i;
	for (i = k+1; i <= n; i++) acc *= i; /* n! div k! */
	for (i = n-k; i > 1; i--) acc /= i;  /* div (n-k)! */
	return acc;
}

static inline double 
v_mag(unsigned n, double const *v /* [n] */)
{
	double acc = 0.0;
	unsigned i;
	for (i = 0; i < n; i++) acc = hypot(acc, v[i]);
	return acc;
}

static inline void 
v_add(unsigned n, double *dst /* [n] */, double *const src /* [n] */)
{
	unsigned i;
	for (i = 0; i < n; i++) dst[i] += src[i];
}

static inline void 
v_s_mul(unsigned n, double *v /* [n] */, double scalar)
{
	unsigned i;
	for (i = 0; i < n; i++) v[i] *= scalar;
}

static inline void 
v_norm(unsigned n, double *v /* [n] */)
{
	double const mag = v_mag(n, v);
	v_s_mul(n, v, 1.0 / mag);
}

static inline double 
v_dot(unsigned n, double const *v1 /* [n] */, double const *v2 /* [n] */)
{
	double acc = 0.0;
	unsigned i;
	for (i = 0; i < n; i++) acc += v1[i] * v2[i];
	return acc;
}

static inline double *
v_dup(unsigned n, double const *v /* [n] */)
{
	double *const newv = xnew(double, n);
	memcpy(newv, v, sizeof(double) * n);
	return newv;
}

static inline void 
v_copy(unsigned n, double *dst /* [n] */, double const *src /* [n] */)
{
	memcpy(dst, src, sizeof(double) * n);
}

static inline void 
unit_point_of_selection(double *output /* [n_hyper] */,
			unsigned n_hyper, unsigned char const *bound_class /* [n_hyper] */,
			unsigned selection)
{
	unsigned i;
	for (i = 0; i < n_hyper; i++)
		output[bound_class[i]] = (selection & (1 << i))? +0.5 : -0.5;
}

static void 
rotate_point(unsigned n_hyper,
	     double *out /* [n_hyper] */,
	     double const *in /* [n_hyper] */,
	     double const *rotation /* [n_hyper][n_hyper] */)
{
	double *tmp = (out == in)? alloca(sizeof(double) * n_hyper) : out;
	unsigned i, j;

	for (i = 0; i < n_hyper; i++) {
		double const *const row = &rotation[i * n_hyper];
		tmp[i] = 0.0;
		for (j = 0; j < n_hyper; j++) tmp[i] += in[j] * row[j];
	}

	if (tmp != out) v_copy(n_hyper, out, tmp);
}

static void 
compute_one_bound(double *output /* [n_rest] */,
		  unsigned n_hyper, double const *rotation /* [n_hyper][n_hyper] */, unsigned n_parallel,
		  unsigned char const *bound_class /* [n_hyper] */)
{
	static double const EPSILON_DISTANCE = 0.000001;
	unsigned const n_rest = n_parallel + 1;
	unsigned const n_proj = n_hyper - n_rest;
	double *const temp_matrix = alloca(sizeof(double) * n_rest * n_rest);
	double *const normal = alloca(sizeof(double) * n_rest);
	double *const point = alloca(sizeof(double) * n_hyper);
	unsigned const select_max = 1 << n_hyper;
	unsigned const plane_incr = 1 << n_parallel;
	unsigned const plane_point_mask = plane_incr - 1;
	double max_distance = 0.0;
	unsigned plane_select;

	/* We have a class of parallel hyperplanes; each hyperplane is determined by a vector values for the
	   perpendicular axes, and contains points corresponding to the vectors of values for the parallel
	   axes.  *_select are bitmasks labeling points such that the n-th bit corresponds to the positiveness
	   of the value of the axis at bound_class[n]; this means that adding plane_incr moves to the next
	   hyperplane and adding 1 moves to the next point.
	*/

	/* The first thing to do is find the normalized projected normal for this class of hyperplanes.  We
	   iterate hyperplanes until we find one whose projected normal has a magnitude greater than
	   EPSILON_DISTANCE, then normalize that.  In order to find the normal, we take the all-negative
	   point, plus the points with one positive along each parallel axis, and then solve for the normal.
	*/

	for (plane_select = 0; plane_select < select_max; plane_select += plane_incr) {
		double distance;
		unsigned i;

		for (i = 0; i < n_rest; i++) {
			unsigned const point_select = plane_select + ((1 << i) & plane_point_mask);
			unit_point_of_selection(point, n_hyper, bound_class, point_select);
			rotate_point(n_hyper, point, point, rotation);
			memcpy(&temp_matrix[i * n_rest], &point[n_proj], sizeof(double) * n_rest);
		}

		for (i = 0; i < n_rest; i++) normal[i] = 1.0;
		solve_linear(n_rest, temp_matrix, normal);
		distance = 1.0 / v_mag(n_rest, normal);

		if (distance > EPSILON_DISTANCE) {
			v_s_mul(n_rest, normal, distance);
			break;
		}
	}

	if (plane_select == select_max) {
		/* We iterated through all of them and they were all below epsilon.  That's not supposed
		   to be possible (we think).  Buggo. */
		abort();
	}

	/* Now we have the normalized, projected normal.  Now we get to find the maximum distance from the
	   origin by iterating one point of each hyperplane and dotting it with the normal. */

	for (plane_select = 0; plane_select < select_max; plane_select += plane_incr) {
		double distance;
		unit_point_of_selection(point, n_hyper, bound_class, plane_select);
		rotate_point(n_hyper, point, point, rotation);
		distance = v_dot(n_rest, &point[n_proj], normal);
		if (distance > max_distance) max_distance = distance;
	}

	/* Now we have the maximum distance from the origin, so the normal divided by that gets to be
	   the vector with which to dot things to see whether they're in the bound for this class. */
	v_copy(n_rest, output, normal);
	v_s_mul(n_rest, output, 1.0 / max_distance);
}

/* BUGGO: make this iterative */
static double *
compute_bounds_rec(double *output /* fill pointer */,
		   unsigned n_hyper, double const *rotation /* [n_hyper][n_hyper] */, unsigned n_parallel,
		   unsigned char *bound_class /* [n_hyper] */, unsigned n_axes_done /* <= n_parallel */)
{
	if (n_axes_done == n_parallel) {
		/* We've chosen parallel axes.  Now, the rest get to be the perpendicular axes. */
		unsigned char *parallel = &bound_class[0];
		unsigned char *perpendicular = &bound_class[n_parallel];
		unsigned axis;

		for (axis = 0; axis < n_hyper; axis++) {
			if (*parallel == axis)
				parallel++;
			else
				*(perpendicular++) = axis;
		}

		compute_one_bound(output, n_hyper, rotation, n_parallel, bound_class);
		return output + (n_parallel + 1);
	} else {
		/* Choose the next parallel axis. */
		unsigned const lower_bound = (n_axes_done == 0)? 0 : (bound_class[n_axes_done-1] + 1);
		unsigned const upper_bound = n_hyper - (n_parallel - n_axes_done);
		unsigned axis;
 
		for (axis = lower_bound; axis <= upper_bound; axis++) {
			bound_class[n_axes_done] = axis;
			output = compute_bounds_rec(output, n_hyper, rotation, n_parallel,
						    bound_class, n_axes_done+1);
		}

		return output;
	}
}

static double *
compute_bounds(unsigned *n_bounds_p /* scalar output */,
	       unsigned n_hyper, double const *rotation /* [n_hyper][n_hyper] */,
	       unsigned n_proj)
{
	unsigned const n_rest = n_hyper - n_proj;
	unsigned const n_bounds = choose(n_hyper, n_rest - 1);
	double *const bounds = xnew(double, n_bounds * n_rest);
	unsigned char *const bound_class = alloca(sizeof(unsigned char) * n_hyper);

	/* bound_class has a weird layout here.  Each bound plane vector will come from resolving a class of
	   parallel bound planes, and bound_class specifies such a class.  Each such class will correspond to
	   a choice of a combination of (n_rest - 1) parallel axes to the hyperplanes in the class (out of the
	   n_hyper available axes), forcing the remaining (n_proj + 1) perpendicular axes.  The indices of the
	   parallel axes, sorted, form the first (n_rest - 1) elements of bound_class, and the indices of the
	   perpendicular axes, sorted, form the remaining (n_proj + 1) elements.
	 */

	compute_bounds_rec(bounds, n_hyper, rotation, n_rest - 1, bound_class, 0);
	*n_bounds_p = n_bounds;
	return bounds;
}

static void 
make_wirec(qc_state_s *qc)
{
	unsigned aa;
	qc->wirec = glGenLists(1);
	glNewList(qc->wirec, GL_COMPILE);
	glBegin(GL_LINE_LOOP);
	for (aa = 0; aa < 40; aa++)
		glVertex2d(cos((double)aa * 0.25 * 0.62831853), sin((double)aa * 0.25 * 0.62831853));
	glEnd();
	glEndList();
}

static void 
make_solidc(qc_state_s *qc)
{
	unsigned aa;
	qc->solidc = glGenLists(1);
	glNewList(qc->solidc, GL_COMPILE);
	glBegin(GL_TRIANGLE_FAN);
	glVertex2d(0.0, 0.0);
	glVertex2d(1.0, 0.0);
	for (aa = 1; aa < 40; aa++)
		glVertex2d(cos((double)aa * 0.25 * 0.62831853), sin((double)aa * 0.25 * 0.62831853));
	glVertex2d(1.0, 0.0);
	glEnd();
	glEndList();
}

static void 
init_with_matrices(qc_state_s *qc,
		   unsigned n_hyper,
		   double const *rotation /* [n_hyper][n_hyper] */,
		   double const *offset /* [n_hyper] */)
{
	unsigned const n_proj = 2;
	unsigned const n_rest = n_hyper - n_proj;

	qc->n_hyper = n_hyper;
	qc->n_proj = n_proj;
	qc->n_rest = n_rest;
	qc->seen_vertices = NULL;
	qc->rotation = v_dup(n_hyper * n_hyper, rotation);
	qc->offset = v_dup(n_hyper, offset);
	qc->bounds = compute_bounds(&qc->n_bounds, n_hyper, rotation, n_proj);
	make_solidc(qc);
	make_wirec(qc);

	qc->draw_type = DRAW_QUAD;
	qc->use_pcolor = 0;
}

static void
do_reshape(qc_state_s *qc, int width, int height)
{
	GLfloat const aspect = (GLfloat)width / (GLfloat)height;

	glViewport(0, 0, width, height);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	if (aspect >= 1.0) {
		gluOrtho2D(-aspect, +aspect, -1.0f, +1.0f);
	} else {
		GLfloat const inv_aspect = 1.0f / aspect;
		gluOrtho2D(-1.0f, +1.0f, -inv_aspect, +inv_aspect);
	}

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glClearColor(0.f, 0.f, 0.f, 0.f);
	glClear(GL_COLOR_BUFFER_BIT);
}

static int 
test_bounds(unsigned n, double const *point /* [n] */,
	    unsigned n_bounds, double const *bounds /* [n * n_bounds] */)
{
	while (n_bounds--) {
		if (fabs(v_dot(n, point, bounds)) >= 0.99999) return 0;
		bounds += n;
	}

	return 1;
}

static int 
seen_vertex_test(qc_state_s *qc, int *pt)
{
	for (vertex_s *v = qc->seen_vertices; v; v = v->next) {
		if (!memcmp(v->coord, pt, sizeof(int) * qc->n_hyper))
			return 1;
	}

	vertex_s *const nv = xnew(vertex_s, 1);
	nv->next = qc->seen_vertices;
	nv->coord = xnew(int, qc->n_hyper);
	memcpy(nv->coord, pt, sizeof(int) * qc->n_hyper);
	qc->seen_vertices = nv;
	return 0;
}

static void 
unsee_vertices(qc_state_s *qc)
{
	for (vertex_s *v = qc->seen_vertices, *w = NULL; v; v = w) {
		w = v->next;
		free(v->coord);
		free(v);
	}

	qc->seen_vertices = NULL;
}

static void 
set_color_from_parity(qc_state_s *qc, unsigned odd)
{
	switch (odd) {
	    case 0: glColor3f(0.5f, 0.1f, 0.1f); break;
	    case 1: glColor3f(0.2f, 0.2f, 0.9f); break;
	    case 2: glColor3f(0.2f, 0.2f, 0.2f); break;
	    case 3: glColor3f(0.7f, 0.7f, 0.7f); break;
	}
}

static void 
draw_tile_ellipse(qc_state_s *qc, unsigned odd, double *points[5])
{
	GLdouble m[16];

	glPushMatrix();
	m[0] = (points[3][0] - points[4][0]) / 2;
	m[1] = (points[3][1] - points[4][1]) / 2;
	m[2] = 0.0;
	m[3] = 0.0;
	m[4] = (points[2][0] - points[1][0]) / 2;
	m[5] = (points[2][1] - points[1][1]) / 2;
	m[6] = 0.0;
	m[7] = 0.0;
	m[8] = 0.0;
	m[9] = 0.0;
	m[10] = 1.0;
	m[11] = 0.0;
	m[12] = (points[3][0] + points[4][0]) / 2;
	m[13] = (points[3][1] + points[4][1]) / 2;
	m[14] = 0.0;
	m[15] = 1.0;
	glMultMatrixd(m);

	set_color_from_parity(qc, odd);
	glCallList(qc->wirec);
	glPopMatrix();
}

static void 
draw_tile_solid(qc_state_s *qc, unsigned selection, double *points[5])
{
	set_color_from_parity(qc, selection);
	glBegin(GL_QUADS);
	glVertex2d(points[4][0], points[4][1]);
	glVertex2d(points[1][0], points[1][1]);
	glVertex2d(points[3][0], points[3][1]);
	glVertex2d(points[2][0], points[2][1]);
	glEnd();

	glColor3f(1.f, 1.f, 1.f);
	glBegin(GL_LINE_LOOP);
	glVertex2d(points[4][0], points[4][1]);
	glVertex2d(points[1][0], points[1][1]);
	glVertex2d(points[3][0], points[3][1]);
	glVertex2d(points[2][0], points[2][1]);
	glEnd();
}

static void 
draw_tile_dual_pcolor(qc_state_s *qc, unsigned selection, double *points[5])
{
	set_color_from_parity(qc, selection);
	glBegin(GL_LINES);
	glVertex2d(points[4][0], points[4][1]);
	glVertex2d(points[3][0], points[3][1]);
	glVertex2d(points[1][0], points[1][1]);
	glVertex2d(points[2][0], points[2][1]);
	glEnd();
}

static void 
draw_tile_dual_vcolor(qc_state_s *qc, unsigned odd, double *points[5])
{
	glBegin(GL_LINES);
	set_color_from_parity(qc, odd);
	glVertex2d(points[4][0], points[4][1]);
	glVertex2d(points[3][0], points[3][1]);
	set_color_from_parity(qc, odd^1);
	glVertex2d(points[1][0], points[1][1]);
	glVertex2d(points[2][0], points[2][1]);
	glEnd();
}

static void 
maybe_draw_tile(qc_state_s *qc, unsigned i, unsigned j, int *pt, double *points[5])
{
	unsigned const n_hyper = qc->n_hyper;
	unsigned const n_rest = qc->n_rest;
	unsigned const n_proj = qc->n_proj;
	unsigned const thin = ((j - i) == 2 || (j - i) == 3);
	unsigned odd = 0;
	for (unsigned z = 0; z < n_hyper; z++) odd ^= (pt[z] & 1);

	for (unsigned k = 1; k < 4; k++) {
		rotate_point(n_hyper, points[k], points[k], qc->rotation);
		if (!test_bounds(n_rest, points[k] + n_proj, qc->n_bounds, qc->bounds))
			return;
	}

	switch (qc->draw_type) {
	    case DRAW_QUAD: draw_tile_solid(qc, qc->use_pcolor? thin : 2, points); break;
	    case DRAW_ELLIPSE: draw_tile_ellipse(qc, qc->use_pcolor? thin : 3, points); break;
	    case DRAW_DUAL_VCOLOR: draw_tile_dual_vcolor(qc, odd, points); break;
	    case DRAW_DUAL_PCOLOR: draw_tile_dual_pcolor(qc, qc->use_pcolor? thin : 3, points); break;
	}
}

static void 
recurse(qc_state_s *qc, int *pt, unsigned depth)
{
	if (depth == 0) return;
	if (seen_vertex_test(qc, pt)) return;

	unsigned const n_hyper = qc->n_hyper;
	unsigned const n_rest = qc->n_rest;
	unsigned const n_proj = qc->n_proj;
	
	double *const points_storage = alloca(sizeof(double) * n_hyper * 5);
	double *const points[5] = {
		&points_storage[0],
		&points_storage[n_hyper],
		&points_storage[2 * n_hyper],
		&points_storage[3 * n_hyper],
		&points_storage[4 * n_hyper],
	};

	unsigned i, j, k;

	memset(points_storage, 0, sizeof(double) * n_hyper * 4);
	for (k = 0; k < n_hyper; k++) points[0][k] = pt[k];
	v_add(n_hyper, points[0], qc->offset);
	rotate_point(n_hyper, points[4], points[0], qc->rotation);
	if (!test_bounds(n_rest, points[4] + n_proj, qc->n_bounds, qc->bounds))
		return;

	for (i = 0; i < n_hyper-1; i++) {
		for (j = i+1; j < n_hyper; j++) {
			v_copy(n_hyper, points[1], points[0]);
			v_copy(n_hyper, points[2], points[0]);
			v_copy(n_hyper, points[3], points[0]);
			points[1][i] += 1.0; points[3][i] += 1.0;
			points[2][j] += 1.0; points[3][j] += 1.0;

			maybe_draw_tile(qc, i, j, pt, points);
		}
	}

	for (i = 0; i < n_hyper; i++) {
		pt[i]++; recurse(qc, pt, depth-1); pt[i]--;
		pt[i]--; recurse(qc, pt, depth-1); pt[i]++;
	}
}

static void
do_draw(qc_state_s *qc)
{
	int *const pt = alloca(sizeof(int) * qc->n_hyper);
	memset(pt, 0, sizeof(int) * qc->n_hyper);

	glClear(GL_COLOR_BUFFER_BIT);

	glPushMatrix();
	glScalef(0.2f, 0.2f, 1.0f);
	glColor3f(1.f, 1.f, 1.f);

	recurse(qc, pt, 50);
	unsee_vertices(qc);

	glPopMatrix();
}

static void 
do_init(qc_state_s *qc)
{
	double const sigma = (1.0 - sqrt(5.0)) / 2.0;
	double const tau = (1.0 + sqrt(5.0)) / 2.0;
	double const beta = sqrt(3.0 - tau);
	double const tau_beta = tau*beta;
	double const sqrt2 = sqrt(2.0);
	double rotation[25] = {
		2.0, -sigma, -tau, -tau, -sigma,
		0.0, tau_beta, beta, -beta, -tau_beta,
		2.0, -tau, -sigma, -sigma, -tau,
		0.0, beta, -tau_beta, tau_beta, -beta,
		sqrt2, sqrt2, sqrt2, sqrt2, sqrt2
	};
	double const offset[5] = { 0.0, 0.3, -0.45, -0.1, 0.0 };
	v_s_mul(25, rotation, sqrt(0.1));
	init_with_matrices(qc, 5, rotation, offset);

	do_reshape(qc, WIDTH, HEIGHT);
}

static void 
do_event(qc_state_s *qc, SDL_Event *ev)
{
	if (ev->type == SDL_KEYDOWN) {
		switch (ev->key.keysym.sym) {
		    case SDLK_1: qc->draw_type = DRAW_QUAD; break;
		    case SDLK_2: qc->draw_type = DRAW_ELLIPSE; break;
		    case SDLK_3: qc->draw_type = DRAW_DUAL_PCOLOR; break;
		    case SDLK_4: qc->draw_type = DRAW_DUAL_VCOLOR; break;
		    case SDLK_c: qc->use_pcolor = !qc->use_pcolor; break;
		    default: ;
		}
	}
}

int 
main(int argc, char *argv[])
{
	/* BUGGO: check for init errors */
	SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	SDL_SetVideoMode(WIDTH, HEIGHT, 0, SDL_OPENGL | SDL_FULLSCREEN);
	SDL_ShowCursor(SDL_DISABLE);

	static qc_state_s qc;
	do_init(&qc);

	Uint32 last_frame = SDL_GetTicks();

	for (;;) {
		SDL_Event ev;

		while (SDL_PollEvent(&ev)) {
			if (ev.type == SDL_QUIT || (ev.type == SDL_KEYDOWN && ev.key.keysym.sym == SDLK_q))
				goto quit;
			do_event(&qc, &ev);
		}

		Uint32 const now = SDL_GetTicks();
		do_draw(&qc);
		last_frame = now;
		SDL_GL_SwapBuffers();
		SDL_Delay(5);
	}
quit:
	SDL_Quit();
	return 0;
}

/* Local variables: */
/* mode: c */
/* comment-column: 40 */
/* fill-column: 110 */
/* End: */
